from __future__ import annotations

import inspect
from functools import lru_cache
from typing import Any, Mapping, Sequence, Union

import logging

from . import AsyncSession, Session
from .model import _model_columns, _single_pk_name

logger = logging.getLogger("uvicorn")


@lru_cache(maxsize=512)
def _is_async_db_type(db_type: type[Any]) -> bool:
    return db_type.__name__ == "AsyncSession" or hasattr(db_type, "run_sync")


def _is_async_db(db: Any) -> bool:
    logger.debug("_is_async_db called with db=%s", db)
    result = _is_async_db_type(type(db))
    logger.debug("_is_async_db returning %s", result)
    return result


async def _maybe_get(db: Union[Session, AsyncSession], model: type, pk_value: Any):
    logger.debug("_maybe_get model=%s pk_value=%s", model, pk_value)
    result = db.get(model, pk_value)  # type: ignore[attr-defined]
    if inspect.isawaitable(result):
        result = await result
    logger.debug("_maybe_get returning %s", result)
    return result


async def _maybe_execute(db: Union[Session, AsyncSession], stmt: Any):
    logger.debug("_maybe_execute stmt=%s", stmt)
    result = db.execute(stmt)  # type: ignore[attr-defined]
    if inspect.isawaitable(result):
        result = await result
    logger.debug("_maybe_execute returning %s", result)
    return result


async def _maybe_flush(db: Union[Session, AsyncSession]) -> None:
    logger.debug("_maybe_flush called")
    result = db.flush()  # type: ignore[attr-defined]
    if inspect.isawaitable(result):
        await result
    logger.debug("_maybe_flush completed")


async def _maybe_rollback(db: Union[Session, AsyncSession]) -> None:
    logger.debug("_maybe_rollback called")
    if not hasattr(db, "rollback"):
        logger.debug("_maybe_rollback skipping rollback; no attribute")
        return
    result = db.rollback()  # type: ignore[attr-defined]
    if inspect.isawaitable(result):
        await result
    logger.debug("_maybe_rollback completed")


async def _maybe_delete(db: Union[Session, AsyncSession], obj: Any) -> None:
    logger.debug("_maybe_delete called with obj=%s", obj)
    if not hasattr(db, "delete"):
        logger.debug("_maybe_delete skipping delete; no attribute")
        return
    result = db.delete(obj)  # type: ignore[attr-defined]
    if inspect.isawaitable(result):
        await result
    logger.debug("_maybe_delete completed for obj=%s", obj)


def _set_attrs(
    obj: Any,
    values: Mapping[str, Any],
    *,
    allow_missing: bool = True,
    skip: Sequence[str] = (),
) -> None:
    logger.debug(
        "_set_attrs called on obj=%s values=%s allow_missing=%s skip=%s",
        obj,
        values,
        allow_missing,
        skip,
    )
    cols = set(_model_columns(type(obj)))
    pk = _single_pk_name(type(obj))
    skip_set = set(skip) | {pk}

    if allow_missing:
        for k, v in values.items():
            if k in cols and k not in skip_set:
                setattr(obj, k, v)
    else:
        for c in cols:
            if c in skip_set:
                continue
            if c in values:
                setattr(obj, c, values[c])
            else:
                setattr(obj, c, None)
    logger.debug("_set_attrs completed for obj=%s", obj)



from __future__ import annotations

from typing import Any, Dict, Mapping, Optional, Union
import builtins as _builtins
import logging

from . import AsyncSession, Session

logger = logging.getLogger("uvicorn")


def _pop_bound_self(args: list[Any]) -> None:
    logger.debug("_pop_bound_self called with args=%s", args)
    if args and not isinstance(args[0], type):
        args.pop(0)
    logger.debug("_pop_bound_self result args=%s", args)


def _extract_db(
    args: list[Any], kwargs: dict[str, Any]
) -> Union[Session, AsyncSession]:
    logger.debug("_extract_db called with args=%s kwargs=%s", args, kwargs)
    db = kwargs.pop("db", None)
    if db is not None:
        logger.debug("_extract_db found db in kwargs=%s", db)
        return db
    for i, a in enumerate(args):
        if isinstance(a, (Session, AsyncSession)) or hasattr(a, "execute"):
            args.pop(i)
            logger.debug("_extract_db using positional db=%s", a)
            return a  # type: ignore[return-value]
    logger.debug("_extract_db failed to find db")
    raise TypeError("db session is required")


def _as_pos_int(x: Any) -> Optional[int]:
    logger.debug("_as_pos_int called with x=%s", x)
    if x is None:
        return None
    try:
        v = int(x)
        result = v if v >= 0 else 0
        logger.debug("_as_pos_int returning %s", result)
        return result
    except Exception:
        logger.debug("_as_pos_int returning None for x=%s", x)
        return None


def _normalize_list_call(
    _args: tuple[Any, ...], _kwargs: dict[str, Any]
) -> tuple[type, Dict[str, Any]]:
    logger.debug("_normalize_list_call called with _args=%s _kwargs=%s", _args, _kwargs)
    args = _builtins.list(_args)
    kwargs = dict(_kwargs)

    _pop_bound_self(args)

    if args and isinstance(args[0], type):
        model = args.pop(0)
    else:
        model = kwargs.pop("model", None)
        if not isinstance(model, type):
            raise TypeError("list(model, ...) requires a model class")

    filters = kwargs.pop("filters", None)
    if filters is None and args:
        maybe = args[0]
        if isinstance(maybe, Mapping):
            filters = args.pop(0)

    skip = _as_pos_int(kwargs.pop("skip", None))
    limit = _as_pos_int(kwargs.pop("limit", None))
    sort = kwargs.pop("sort", None)

    if skip is None and args:
        skip = _as_pos_int(args[0])
        if skip is not None:
            args.pop(0)
    if limit is None and args:
        limit = _as_pos_int(args[0])
        if limit is not None:
            args.pop(0)

    db = _extract_db(args, kwargs)

    if filters is None:
        filters = {}

    result = {
        "filters": filters,
        "skip": skip,
        "limit": limit,
        "db": db,
        "sort": sort,
    }
    logger.debug("_normalize_list_call returning model=%s params=%s", model, result)
    return model, result



from __future__ import annotations

try:
    from sqlalchemy import select, delete as sa_delete, and_, asc, desc, Enum as SAEnum
    from sqlalchemy.orm import Session
    from sqlalchemy.ext.asyncio import AsyncSession
    from sqlalchemy.orm.exc import NoResultFound  # type: ignore
except Exception:  # pragma: no cover
    select = sa_delete = and_ = asc = desc = None  # type: ignore
    SAEnum = None  # type: ignore
    Session = object  # type: ignore
    AsyncSession = object  # type: ignore

    class NoResultFound(LookupError):  # type: ignore
        pass


from .model import (
    _pk_columns,
    _single_pk_name,
    _coerce_pk_value,
    _model_columns,
    _colspecs,
    _filter_in_values,
    _immutable_columns,
)
from .filters import _CANON_OPS, _coerce_filters, _apply_filters, _apply_sort
from .db import (
    _is_async_db,
    _maybe_get,
    _maybe_execute,
    _maybe_flush,
    _maybe_rollback,
    _maybe_delete,
    _set_attrs,
)
from .enum import _validate_enum_values
from .normalize import (
    _normalize_list_call,
    _pop_bound_self,
    _extract_db,
    _as_pos_int,
)

__all__ = [
    "AsyncSession",
    "Session",
    "NoResultFound",
    "select",
    "sa_delete",
    "_apply_filters",
    "_apply_sort",
    "_CANON_OPS",
    "_coerce_filters",
    "_coerce_pk_value",
    "_colspecs",
    "_filter_in_values",
    "_immutable_columns",
    "_is_async_db",
    "_maybe_delete",
    "_maybe_execute",
    "_maybe_flush",
    "_maybe_rollback",
    "_maybe_get",
    "_model_columns",
    "_normalize_list_call",
    "_pop_bound_self",
    "_extract_db",
    "_as_pos_int",
    "_pk_columns",
    "_set_attrs",
    "_single_pk_name",
    "_validate_enum_values",
    "SAEnum",
    "asc",
    "desc",
    "and_",
]



"""OLTP operation implementations for Tigrbl."""

from .crud import (  # noqa: F401
    Body,
    Header,
    Param,
    Path,
    Query,
    bulk_create,
    bulk_delete,
    bulk_merge,
    bulk_replace,
    bulk_update,
    clear,
    create,
    delete,
    list,
    merge,
    read,
    replace,
    update,
)



from __future__ import annotations

import enum as _enum
from typing import Any, Mapping
import builtins as _builtins
import logging

from . import SAEnum

logger = logging.getLogger("uvicorn")


def _validate_enum_values(model: type, values: Mapping[str, Any]) -> None:
    logger.debug("_validate_enum_values called with model=%s values=%s", model, values)
    if not values or SAEnum is None:
        logger.debug("_validate_enum_values no validation needed")
        return

    table = getattr(model, "__table__", None)
    if table is None:
        return

    get = getattr(table.c, "get", None)

    for key, v in values.items():
        col = get(key) if get else None
        if col is None:
            try:
                col = table.c[key]  # type: ignore[index]
            except Exception:
                col = None
        if col is None:
            continue

        col_type = getattr(col, "type", None)
        if col_type is None or not isinstance(col_type, SAEnum):
            continue

        if v is None:
            continue

        enum_cls = getattr(col_type, "enum_class", None)
        if enum_cls is not None:
            if isinstance(v, _enum.Enum):
                if isinstance(v, enum_cls):
                    continue
                logger.debug(
                    "_validate_enum_values invalid value %s for enum %s", v, enum_cls
                )
                raise LookupError(
                    f"{v!r} is not among the defined enum values. "
                    f"Enum name: {enum_cls.__name__}. "
                    f"Possible values: {', '.join([e.value for e in enum_cls])}"
                )

            allowed_values = [e.value for e in enum_cls]
            allowed_names = [e.name for e in enum_cls]
            if isinstance(v, str) and (v in allowed_values or v in allowed_names):
                continue

            logger.debug(
                "_validate_enum_values invalid value %s for enum %s", v, enum_cls
            )
            raise LookupError(
                f"{v!r} is not among the defined enum values. "
                f"Enum name: {enum_cls.__name__}. "
                f"Possible values: {', '.join(allowed_values)}"
            )
        else:
            allowed = _builtins.list(getattr(col_type, "enums", []) or [])
            if isinstance(v, str) and v in allowed:
                continue
            logger.debug(
                "_validate_enum_values invalid value %s for enum %s", v, col_type
            )
            raise LookupError(
                f"{v!r} is not among the defined enum values. "
                f"Enum name: {getattr(col_type, 'name', 'Enum')}. "
                f"Possible values: {', '.join(allowed) if allowed else '(none)'}"
            )
    logger.debug("_validate_enum_values completed")



from __future__ import annotations

from functools import lru_cache
from typing import Any, Dict, List, Mapping, Optional, Union
import inspect
from weakref import WeakSet

import builtins as _builtins
import logging

from sqlalchemy.orm.exc import UnmappedInstanceError
from sqlalchemy import inspect as _sa_inspect
from sqlalchemy.exc import NoInspectionAvailable, OperationalError

from .helpers import (
    AsyncSession,
    Session,
    NoResultFound,
    select,
    sa_delete,
    _apply_filters,
    _apply_sort,
    _coerce_filters,
    _coerce_pk_value,
    _filter_in_values,
    _immutable_columns,
    _maybe_delete,
    _maybe_execute,
    _maybe_flush,
    _maybe_get,
    _normalize_list_call,
    _set_attrs,
    _single_pk_name,
    _validate_enum_values,
)

logger = logging.getLogger("uvicorn")

_MAPPED_MODELS: "WeakSet[type]" = WeakSet()


@lru_cache(maxsize=512)
def _create_flush_requirements(model: type) -> tuple[str | None, tuple[str, ...]]:
    """Precompute model attributes that may need a create-time flush."""
    table = getattr(model, "__table__", None)
    pk_name: str | None = None
    server_default_names: list[str] = []

    if table is None:
        return pk_name, ()

    try:
        pk_cols = tuple(getattr(table, "primary_key", ()).columns)
    except Exception:
        pk_cols = ()
    if len(pk_cols) == 1:
        name = getattr(pk_cols[0], "name", None)
        if isinstance(name, str) and name:
            pk_name = name

    for col in getattr(table, "columns", ()):
        name = getattr(col, "name", None)
        if not isinstance(name, str) or not name:
            continue
        if getattr(col, "server_default", None) is not None:
            server_default_names.append(name)

    return pk_name, tuple(server_default_names)


def _requires_create_flush(model: type, obj: Any) -> bool:
    """Flush only when we need database-generated values immediately."""
    pk_name, server_default_names = _create_flush_requirements(model)

    if pk_name is not None and getattr(obj, pk_name, None) is None:
        return True

    for name in server_default_names:
        if getattr(obj, name, None) is None:
            return True
    return False


def _ensure_model_mapped(model: type) -> None:
    if model in _MAPPED_MODELS:
        return
    try:
        _sa_inspect(model)
        _MAPPED_MODELS.add(model)
        return
    except NoInspectionAvailable:
        pass

    from tigrbl_orm.orm.tables import TableBase
    from tigrbl_base._base._table_base import _materialize_colspecs_to_sqla

    _materialize_colspecs_to_sqla(model)
    TableBase.registry.map_declaratively(model)
    _MAPPED_MODELS.add(model)


def _ensure_model_table(model: type, db: Union[Session, AsyncSession]) -> None:
    """Best-effort table creation for in-memory sqlite race conditions."""
    table = getattr(model, "__table__", None)
    if table is None:
        return
    bind = getattr(db, "bind", None)
    if bind is None and hasattr(db, "get_bind"):
        try:
            bind = db.get_bind()
        except Exception:
            bind = None
    if bind is None:
        return
    try:
        table.create(bind=bind, checkfirst=True)
    except Exception:
        return


async def create(
    model: type, data: Mapping[str, Any], db: Union[Session, AsyncSession]
) -> Any:
    """
    Insert a single row. Returns the persisted model instance.
    Flush-only (commit happens later in END_TX).
    """
    logger.debug("create called with model=%s data=%s", model, data)
    _ensure_model_mapped(model)
    data = _filter_in_values(model, data or {}, "create")
    _validate_enum_values(model, data)
    obj = model(**data)
    if hasattr(db, "add"):
        try:
            db.add(obj)
        except UnmappedInstanceError:
            # Test suites may dispose the declarative registry between app
            # instances. Re-materialize and re-map before retrying create.
            from tigrbl_orm.orm.tables import TableBase
            from tigrbl_base._base._table_base import _materialize_colspecs_to_sqla

            _materialize_colspecs_to_sqla(model)
            TableBase.registry.map_declaratively(model)
            obj = model(**data)
            db.add(obj)
    needs_flush = _requires_create_flush(model, obj)

    try:
        if needs_flush:
            await _maybe_flush(db)
    except OperationalError as exc:
        if "no such table" not in str(exc).lower():
            raise
        if hasattr(db, "rollback"):
            rollback_result = db.rollback()
            if inspect.isawaitable(rollback_result):
                await rollback_result
        _ensure_model_table(model, db)
        obj = model(**data)
        db.add(obj)
        if needs_flush:
            await _maybe_flush(db)
    logger.debug("create persisted obj=%s", obj)
    return obj


async def read(model: type, ident: Any, db: Union[Session, AsyncSession]) -> Any:
    """
    Load a single row by primary key. Raises NoResultFound if not found.
    """
    logger.debug("read called with model=%s ident=%s", model, ident)
    _ensure_model_mapped(model)
    obj = await _maybe_get(db, model, ident)
    if obj is None:
        logger.debug("read did not find model=%s ident=%s", model, ident)
        raise NoResultFound(f"{model.__name__}({ident!r}) not found")
    logger.debug("read returning obj=%s", obj)
    return obj


async def update(
    model: type, ident: Any, data: Mapping[str, Any], db: Union[Session, AsyncSession]
) -> Any:
    """
    Partial update by primary key. Missing keys are left unchanged.
    Returns the updated model instance. Flush-only.
    """
    logger.debug("update called with model=%s ident=%s data=%s", model, ident, data)
    _ensure_model_mapped(model)
    data = _filter_in_values(model, data or {}, "update")
    _validate_enum_values(model, data)
    obj = await read(model, ident, db)
    skip = _immutable_columns(model, "update")
    _set_attrs(obj, data, allow_missing=True, skip=skip)
    await _maybe_flush(db)
    logger.debug("update returning obj=%s", obj)
    return obj


async def replace(
    model: type, ident: Any, data: Mapping[str, Any], db: Union[Session, AsyncSession]
) -> Any:
    """
    PUT semantics with upsert behaviour.

    If the row exists it is replaced entirely (missing attributes are nulled).
    If the row does not exist it is created with the provided identifier.
    Flush-only.
    """
    logger.debug("replace called with model=%s ident=%s data=%s", model, ident, data)
    _ensure_model_mapped(model)
    data = _filter_in_values(model, data or {}, "replace")
    _validate_enum_values(model, data)
    pk = _single_pk_name(model)
    obj = await _maybe_get(db, model, ident)
    if obj is None:
        payload = {pk: ident, **data}
        result = await create(model, payload, db=db)
        logger.debug("replace created obj=%s", result)
        return result
    skip = _immutable_columns(model, "replace")
    _set_attrs(obj, data, allow_missing=False, skip=skip)
    await _maybe_flush(db)
    logger.debug("replace updated obj=%s", obj)
    return obj


async def merge(
    model: type, ident: Any, data: Mapping[str, Any], db: Union[Session, AsyncSession]
) -> Any:
    """PATCH semantics with upsert behaviour."""
    logger.debug("merge called with model=%s ident=%s data=%s", model, ident, data)
    _ensure_model_mapped(model)
    pk = _single_pk_name(model)
    ident = _coerce_pk_value(model, ident)
    obj = await _maybe_get(db, model, ident)

    verb = "update" if obj is not None else "create"
    data = _filter_in_values(model, data or {}, verb)
    _validate_enum_values(model, data)
    data_no_pk = {k: v for k, v in data.items() if k != pk}
    if obj is None:
        payload = {pk: ident, **data_no_pk}
        result = await create(model, payload, db=db)
        logger.debug("merge created obj=%s", result)
        return result
    skip = _immutable_columns(model, "update")
    _set_attrs(obj, data_no_pk, allow_missing=True, skip=skip)
    await _maybe_flush(db)
    logger.debug("merge updated obj=%s", obj)
    return obj


async def delete(
    model: type, ident: Any, db: Union[Session, AsyncSession]
) -> Dict[str, int]:
    """
    Delete by primary key. Returns {"deleted": 1} if removed, else raises NoResultFound.
    Flush-only.
    """
    logger.debug("delete called with model=%s ident=%s", model, ident)
    _ensure_model_mapped(model)
    obj = await read(model, ident, db)
    await _maybe_delete(db, obj)
    await _maybe_flush(db)
    logger.debug("delete removed obj=%s", obj)
    return {"deleted": 1}


# NOTE: tolerant signature: accepts positional/keyword and ignores stray args
async def list(*_args: Any, **_kwargs: Any) -> List[Any]:  # noqa: A001  (shadow built-in)
    """
    Simple list with equality filters + skip/limit (+ optional sort).
    Tolerant to:
      - missing filters (defaults to {})
      - accidental bound-method 'self' (first positional arg)
      - positional or keyword args
      - stray extras (e.g., request) which are ignored
    """
    logger.debug("list called with args=%s kwargs=%s", _args, _kwargs)
    model, params = _normalize_list_call(_args, _kwargs)
    _ensure_model_mapped(model)

    filters: Mapping[str, Any] = _coerce_filters(model, params["filters"])
    skip: Optional[int] = params["skip"]
    limit: Optional[int] = params["limit"]
    db: Union[Session, AsyncSession] = params["db"]
    sort = params["sort"]

    if select is None:  # pragma: no cover
        # Fallback: legacy query API
        q = db.query(model)  # type: ignore[attr-defined]
        if filters:
            q = q.filter_by(**filters)  # type: ignore[attr-defined]
        if isinstance(skip, int):
            q = q.offset(max(skip, 0))  # type: ignore[attr-defined]
        if isinstance(limit, int) and limit is not None:
            q = q.limit(max(limit, 0))  # type: ignore[attr-defined]
        return _builtins.list(q.all())  # type: ignore[attr-defined]

    where = _apply_filters(model, filters)
    stmt = select(model)
    if where is not None:
        stmt = stmt.where(where)

    order_exprs = _apply_sort(model, sort)
    if order_exprs:
        for ob in order_exprs:
            stmt = stmt.order_by(ob)

    if isinstance(skip, int):
        stmt = stmt.offset(max(skip, 0))
    if isinstance(limit, int) and limit is not None:
        stmt = stmt.limit(max(limit, 0))

    result = await _maybe_execute(db, stmt)
    items = _builtins.list(result.scalars().all())  # type: ignore[attr-defined]
    logger.debug("list returning %d items", len(items))
    return items


async def clear(
    *args: Any,
    **kwargs: Any,
) -> Dict[str, int]:
    """
    Delete many rows matching equality filters. Returns {"deleted": N}.
    Flush-only. Tolerant to the same calling variations as `list`.
    """
    # Reuse normalizer to accept the same shapes
    logger.debug("clear called with args=%s kwargs=%s", args, kwargs)
    model, params = _normalize_list_call(args, kwargs)
    _ensure_model_mapped(model)
    raw_filters: Mapping[str, Any] = params["filters"]
    db: Union[Session, AsyncSession] = params["db"]

    if sa_delete is None:  # pragma: no cover
        # Fallback path: manual iteration
        items = await list(model, raw_filters, db=db)
        n = 0
        for obj in items:
            await _maybe_delete(db, obj)
            n += 1
        await _maybe_flush(db)
        return {"deleted": n}

    filt = _coerce_filters(model, raw_filters)
    where = _apply_filters(model, filt)
    stmt = sa_delete(model)
    if where is not None:
        stmt = stmt.where(where)

    res = await _maybe_execute(db, stmt)
    await _maybe_flush(db)
    n = int(getattr(res, "rowcount", 0) or 0)
    logger.debug("clear removed %d rows", n)
    return {"deleted": n}



from __future__ import annotations

import datetime as dt
from functools import lru_cache
from typing import Any, Dict, Mapping, Tuple

try:
    from tigrbl_core._spec.column_spec import mro_collect_columns
except Exception:  # pragma: no cover

    def mro_collect_columns(_model: type, _cache_bust: int | None = None):
        return {}


def _colspecs_cache_token(model: type) -> tuple[int, int, int, int]:
    table = getattr(model, "__table__", None)
    cols = getattr(table, "columns", None) if table is not None else None
    try:
        col_count = len(tuple(cols)) if cols is not None else 0
    except Exception:
        col_count = 0
    return (
        id(getattr(model, "__tigrbl_colspecs__", None)),
        id(getattr(model, "__tigrbl_cols__", None)),
        id(table),
        col_count,
    )


@lru_cache(maxsize=256)
def _colspecs_cached(
    model: type, cache_token: tuple[int, int, int, int]
) -> Mapping[str, Any]:
    del cache_token
    return mro_collect_columns(model, _cache_bust=hash(_colspecs_cache_token(model)))


@lru_cache(maxsize=512)
def _excluded_schema_fields(model: type, verb: str) -> frozenset[str]:
    schema_cls = getattr(getattr(model, "schemas", None), verb, None)
    schema_in = getattr(schema_cls, "in_", None) if schema_cls else None
    model_fields = (
        getattr(schema_in, "model_fields", None) if schema_in is not None else None
    )
    if not isinstance(model_fields, Mapping):
        return frozenset()
    return frozenset(
        name
        for name, field_info in model_fields.items()
        if getattr(field_info, "exclude", False)
    )


def _filter_plan_cache_token(model: type, verb: str) -> tuple[int, int, int, int, int]:
    return (
        *_colspecs_cache_token(model),
        id(getattr(getattr(model, "schemas", None), verb, None)),
    )


@lru_cache(maxsize=512)
def _filter_plan_cached(
    model: type, verb: str, cache_token: tuple[int, int, int, int, int]
) -> tuple[Mapping[str, bool], frozenset[str], Mapping[str, Any]]:
    del cache_token
    specs = _colspecs(model)
    if not specs:
        return {}, frozenset(), _table_python_types(model)

    allowed_by_field: dict[str, bool] = {}
    for name, sp in specs.items():
        io = getattr(sp, "io", None)
        if io is None:
            allowed_by_field[name] = True
            continue
        in_verbs = getattr(io, "in_verbs", ())
        mutable = getattr(io, "mutable_verbs", ())
        allowed = (not in_verbs or verb in in_verbs) and (
            not mutable or verb in mutable
        )
        allowed_by_field[name] = allowed

    return (
        allowed_by_field,
        _excluded_schema_fields(model, verb),
        _table_python_types(model),
    )


@lru_cache(maxsize=256)
def _table_python_types(model: type) -> Mapping[str, Any]:
    table = getattr(model, "__table__", None)
    columns = getattr(table, "columns", None)
    if columns is None:
        return {}
    resolved: dict[str, Any] = {}
    for col in columns:
        try:
            py_t = getattr(getattr(col, "type", None), "python_type", None)
        except Exception:
            py_t = None
        if py_t is not None:
            resolved[getattr(col, "name", "")] = py_t
    return resolved


def _pk_columns(model: type) -> Tuple[Any, ...]:
    table = getattr(model, "__table__", None)
    if table is None:
        raise ValueError(f"{model.__name__} has no __table__")
    pks = tuple(table.primary_key.columns)  # type: ignore[attr-defined]
    if not pks:
        raise ValueError(f"{model.__name__} has no primary key")
    return pks


def _single_pk_name(model: type) -> str:
    pks = _pk_columns(model)
    if len(pks) != 1:
        raise NotImplementedError(
            f"{model.__name__} has composite PK; not supported by default core"
        )
    name = pks[0].name
    return name


def _coerce_pk_value(model: type, value: Any) -> Any:
    if value is None:
        return None
    try:
        col = _pk_columns(model)[0]
        py_type = col.type.python_type  # type: ignore[attr-defined]
    except Exception:  # pragma: no cover - best effort
        return value
    if isinstance(value, py_type):
        return value
    try:
        coerced = py_type(value)
        return coerced
    except Exception:  # pragma: no cover - fallback to original
        return value


def _model_columns(model: type) -> Tuple[str, ...]:
    table = getattr(model, "__table__", None)
    if table is None:
        return ()
    cols = tuple(c.name for c in table.columns)
    return cols


def _colspecs(model: type) -> Mapping[str, Any]:
    return _colspecs_cached(model, _colspecs_cache_token(model))


def _filter_in_values(
    model: type, data: Mapping[str, Any], verb: str
) -> Dict[str, Any]:
    allowed_by_field, excluded_fields, column_types = _filter_plan_cached(
        model, verb, _filter_plan_cache_token(model, verb)
    )
    if not allowed_by_field and not excluded_fields:
        return dict(data)

    out: Dict[str, Any] = {}
    for k, v in data.items():
        allowed = allowed_by_field.get(k)
        if allowed is None:
            if k in excluded_fields:
                continue
            out[k] = v
            continue
        if allowed:
            try:
                py_t = column_types.get(k)
                if py_t is not None and v is not None and not isinstance(v, py_t):
                    if py_t in (dt.datetime, dt.date) and isinstance(v, str):
                        parsed = py_t.fromisoformat(v)
                        out[k] = parsed
                    else:
                        out[k] = py_t(v)
                else:
                    out[k] = v
            except Exception:
                # Best effort coercion only; preserve original value on failure.
                out[k] = v
    return out


def _immutable_columns(model: type, verb: str) -> set[str]:
    specs = _colspecs(model)
    if not specs:
        return set()
    imm: set[str] = set()
    for name, sp in specs.items():
        io = getattr(sp, "io", None)
        mutable = getattr(io, "mutable_verbs", ()) if io else ()
        if mutable and verb not in mutable:
            imm.add(name)
    return imm



"""Parameter marker helpers for std dependency resolution."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class Param:
    default: Any = None
    alias: str | None = None
    required: bool = False
    location: str = "query"


def Body(default: Any = None, **kwargs: Any) -> Param:
    return Param(
        default=default,
        alias=kwargs.get("alias"),
        required=default is ...,
        location="body",
    )


def Query(default: Any = None, **kwargs: Any) -> Param:
    return Param(
        default=default,
        alias=kwargs.get("alias"),
        required=default is ...,
        location="query",
    )


def Path(default: Any = None, **kwargs: Any) -> Param:
    return Param(
        default=default,
        alias=kwargs.get("alias"),
        required=default is ...,
        location="path",
    )


def Header(default: Any = None, **kwargs: Any) -> Param:
    return Param(
        default=default,
        alias=kwargs.get("alias"),
        required=default is ...,
        location="header",
    )



from .ops import (
    create,
    read,
    update,
    replace,
    merge,
    delete,
    list as _list,
    clear,
)
from .bulk import (
    bulk_create,
    bulk_update,
    bulk_replace,
    bulk_merge,
    bulk_delete,
)

from .params import Body, Header, Param, Path, Query

# Public alias named exactly `list` to preserve API surface
list = _list  # noqa: A001 - intentional shadow of built-in for public API

__all__ = [
    "Header",
    "Path",
    "Query",
    "Body",
    "Param",
    "create",
    "read",
    "update",
    "replace",
    "merge",
    "delete",
    "list",
    "clear",
    "bulk_create",
    "bulk_update",
    "bulk_replace",
    "bulk_merge",
    "bulk_delete",
]
